(* 
compile.ml
Main compiler for the CMajor programming language
by PLT Sandwich
Andrew OReilly, Stephanie Huang
*)

open Ast
open Sast
open Semantics
open Printf

exception Invalid_play of string
exception Illegal_operation of string
exception Type_error of string
exception Duplicate_name of string

exception Not_implemented of string

(* Symbol map *)
module NameMap = Map.Make(String) 

(* Environment: symbol tables for functions, global, local vars *)
type global_environment = {
  variables : (c_type * s_type) NameMap.t;
  functions : func_decl NameMap.t
}

type symbol_table = {
  parent : symbol_table option;
  variables : (c_type * s_type) NameMap.t
}

type env = {
    glob_env : global_environment;
    scope : symbol_table;
    return_type : s_type;
    return_val : c_type;
}

type composition = {
  dur1 : int list;
  dur2 : int list;
  pitches : int array list
}

(* extract value from c_type * s_type tuple *)
let get_value (v, t) = match v with
    | None -> raise Not_found 
    | _ -> v

(* execute play by writing pitches and durations to composition *)
let play s = 
  (* next is a phrase *)
  let handle_phrase next l = match next with Phrase(chords) -> 
    let comp = { dur1 = []; dur2 = []; pitches = [Array.make (Array.length chords) (-1)]}
    in let rec match_length llen alen l =
      if llen >= alen then l else
      match_length (llen + 1) alen ((Array.make (Array.length chords) (-1)) :: l)
    in 
    (* handles the pitches *)
    let (na, new_pitches) = (
      Array.fold_left (fun (i, comp_pitches) chord -> 
        match chord with Chord(p,d) ->
          let new_pitches =
            List.rev (match_length (List.length comp_pitches) (Array.length p) comp_pitches)
          in Array.iteri (fun pidx pitch ->
            (List.nth new_pitches pidx).(i) <- (pitch_to_int pitch)
          ) p; (i + 1, new_pitches)
        | _ -> raise (Type_mismatch ("Error in play"))
      ) (0, comp.pitches) chords)
    (* handles the durations *)
    in (Array.fold_right (fun chord comp -> match chord with Chord(p,d) ->
      (match d with Dur(d1, d2) ->
        {
          dur1 = d1 :: comp.dur1;
          dur2 = d2 :: comp.dur2;
          pitches = new_pitches;
        }
      | _ -> raise (Type_mismatch ("Error in play")))
    | _ -> raise (Type_mismatch ("Error in play"))
    ) chords comp
  ) :: l
  | _ -> raise (Type_mismatch ("Error in play"))
  in match s with Score(phrases) -> 
    Array.fold_right handle_phrase phrases []
  | _ -> raise (Type_mismatch ("Error in play"))

let csv_ints (listarg : int list) = 
  List.fold_right 
   (fun next str -> (string_of_int next) ^ "," ^ str)
   listarg ""

(* Get the value of a variable from a symbol_table *)
let rec find_symtab_var symtab name =
  try
    NameMap.find name symtab.variables
  with Not_found ->
    match symtab.parent with
        Some(parent) -> find_symtab_var parent name
      | _ -> raise Not_found

(* Get the value of a variable from an environment *)
let find_var env name =
  try
    find_symtab_var env.scope name
  with Not_found -> try
    find_symtab_var {
      parent = None;
      variables = env.glob_env.variables
    } name
  with Not_found -> raise Not_found

(* Set the value of a variable within a symbol_table *)
let rec update_var symtab name newval (t : s_type) = 
  try 
    (* update local scope *)
    let (oldval, typ) = NameMap.find name symtab.variables  in
    if t <> typ then raise (Type_mismatch(name))
    else
    (* check array typing *) 
      if t = SArray then if get_arr_type newval <> get_arr_type oldval
        then raise (Type_mismatch(name)) (*else ()*) 
      else (); { 
        parent = symtab.parent;
        variables = NameMap.add name (newval, t) symtab.variables
      }
  with Not_found ->
    match symtab.parent with
        Some(psymtab) ->
          let newparent = update_var psymtab name newval t in
          {
            parent = Some(newparent);
            variables = symtab.variables
          }
      | _ -> raise Not_found

(* Set the value of a variable within an environment *)
let update_symtab env name newval (t : s_type) =
  try 
    let new_symtab = update_var env.scope name newval t in
    newval, {
      glob_env = env.glob_env;
      scope = new_symtab;
      return_type = env.return_type;
      return_val = env.return_val
    }
  with Not_found -> (
    try 
      let new_glob_scope = update_var {
        parent = None;
        variables = env.glob_env.variables
      } name newval t in
      newval, {
        glob_env = {
          variables = new_glob_scope.variables;
          functions = env.glob_env.functions
        };
        scope = env.scope;
        return_type = env.return_type;
        return_val = env.return_val
      }
    with Not_found -> raise Not_found
  )
  | Match_failure(s,l,c) -> raise (Type_error(name ^ c_type_str(newval)))


let rec translate block env = (
  (* translate all expressions to cmaj type *)
    let rec expr expenv = function
        Id(s) -> (
          try 
            let (c,s) = find_var expenv s in
            (c, expenv)
          with Not_found -> raise (Missing_variable ("Error: "^"\""^s^"\""^" not defined!"))
       )
    | Asn(s, e) -> (
      try 
        let (c, env1) = (expr expenv e) in 
        (* cast array to phrase or score as necessary *)
        let cast_check = function
            SPhrase -> (match c with Array(SChord, dat) -> Phrase(dat) | _ -> c)
          | SScore -> (match c with Array(SPhrase, dat) -> Score(dat) | _ -> c)
          | _ -> c
        in let c = cast_check (snd (find_var env1 s)) in
        (* assign value in environment *)
        update_symtab env1 s c (c_to_s_type c)
      with Not_found -> raise (Missing_variable ("Error: "^"\""^s^"\""^" not defined!"))
       )
    | Arr(e) -> (
      let arr_type = c_to_s_type (fst (expr expenv (List.hd e))) in
      let arr_check elem = 
        let c_elem = fst (expr expenv elem) in
        if (is_valid_elem c_elem arr_type) then c_elem
        else raise (Type_error("Error: unexpected type encountered"))
      in let arr_data = Array.of_list (List.map arr_check e) in
      Array(arr_type, arr_data), expenv
    )
    | Arrget(s, e) -> (
      try
        let (arr, env1) = expr expenv (Id(s)) in
        let index = match (fst (expr env1 e)) with
          Int(i) -> i
        | _ -> raise (Type_error ("Error: index value is not an integer"))
        in let arr_data = match arr with 
          Array(typ, dat) -> dat
        | _ -> raise (Type_error ("Error: \""^s^"\" is not an array"))
        in arr_data.(index), env1
      with Invalid_argument x -> raise (Invalid_argument ("index out of bounds"))
    )
    | Arrmod(s, i, e) -> (
      try
        let (arr, env1) = expr expenv (Id(s)) in
        let index = match (fst (expr env1 i)) with
          Int(idx) -> idx
        | _ -> raise (Type_error ("Error: index value is not an integer"))
        in let arr_data = match arr with 
          Array(typ, dat) -> dat
        | _ -> raise (Type_error ("Error: \""^s^"\" is not an array"))
        in let newval = fst (expr env1 e)
        in if is_valid_elem newval (get_arr_type arr) 
          then arr_data.(index) <- newval
        else raise (Type_error ("Error: unexpected type encountered"));
        newval, env1
      with Invalid_argument x -> raise (Invalid_argument ("index out of bounds"))
    )
    | Lit(x) -> (
      match x with
          Intlit(x) -> Int(x), expenv
        | Pitchlit(l,o) -> Pitch(l, o), expenv
        | Tuple(x, y) -> (
          let (ex, env1) = expr expenv x  in
            let (ey, env2) = expr env1 y  in 
            match ex, ey with 
              Int(a), Int(b) -> Dur(a, b), env2
            | Pitch(l, o), Dur(a, b) -> Note(Pitch(l,o), Dur(a,b)), env2
            | Array(SPitch, d), Dur(a, b) -> Chord(d, Dur(a,b)), env2
            | _ -> raise (Type_error("Invalid tuple"))
          )
      )
      | Call(name, elist) -> (
          let argc, argv, callenv = List.fold_left 
            (
              fun (c,v,envb) next -> 
              let lit, enva = expr envb next in
                c + 1, lit :: v, enva
            ) 
            (0, [], expenv) elist 
          in
          match name with
          "play" -> (
            match argv with
            [_ as s] -> (
              let master_score = match s with
                Score(phrases) -> Score(phrases)
              | Phrase(chords) as p -> Score([|p|])
              | Chord(p,d) as c -> Score([|Phrase([|c|])|])
              | Note(p,d) -> Score([|Phrase([|Chord([|p|],d)|])|])
              | _ -> raise (Invalid_play("Invalid call to play"))
              in let comp_list = play master_score in
              let ofile = open_out "out.csv" in
              List.iter (fun comp ->
                let (dur1str, dur2str) = 
                  (csv_ints comp.dur1, csv_ints comp.dur2)
                in List.iter (fun parray ->
                  let pitchstr = csv_ints (Array.to_list parray) in
                  fprintf ofile "%s\n%s\n%s\n" pitchstr dur1str dur2str;
                ) comp.pitches
              ) comp_list;
              close_out ofile;
              Int(1), expenv
            )
            | _ -> raise (Invalid_play("Invalid call to play"))
          )
          | _ -> 
                try
                  let fxn = NameMap.find name callenv.glob_env.functions in
                  let funenv = exec_fun callenv fxn (List.rev argv) in
                  funenv.return_val, {
                    glob_env = funenv.glob_env;
                    scope = callenv.scope;
                    return_type = callenv.return_type;
                    return_val = callenv.return_val
                  }
                with Not_found -> raise (Missing_function("Unknown function " ^ name))
            (*raise (Not_implemented("Custom functions"))*)
      )
    | Binop (e1, op, e2) -> (
      let (lit1, env1) = expr expenv e1 in
      let (lit2, binenv) = expr env1 e2 in
      match op with
        Add -> (
          match (lit1, lit2) with
            Int(x), Int(y) -> Int(x + y), binenv
          | (Pitch(l,o) as p), (Int(y) as i)
          | (Int(y) as i), (Pitch(l,o) as p) -> raise_pitch i p, binenv
          | Dur(x,y), Int(z)
          | Int(z), Dur(x,y) -> dur_add (Dur(z,1)) (Dur(x,y)), binenv
          | Dur(x,y), Dur(z,w) -> dur_add (Dur(x,y)) (Dur(z,w)), binenv
          (* as concatenation *)
          | Note(p1,d1), Note(p2,d2) -> Phrase([| Chord([|p1|],d1); Chord([|p2|],d2) |]), binenv
          | Note(p,d), (Chord(p2,d2) as c) -> Phrase([| Chord([|p|], d); c |]), binenv
          | (Chord(p2,d2) as c), Note(p,d) -> Phrase([| c; Chord([|p|], d) |]), binenv
          | Note(p,d), Phrase(c) -> Phrase(Array.append [| Chord([|p|],d) |] c), binenv
          | Phrase(c), Note(p,d) -> Phrase(Array.append c [| Chord([|p|],d) |]), binenv
          | Chord(p1,d1) as c1, (Chord(p2,d2) as c2) -> Phrase([| c1; c2 |]), binenv
          | Phrase(c), (Chord(p,d) as ch) -> Phrase(Array.append c [|ch|]), binenv
          | Chord(p,d) as ch, Phrase(c) -> Phrase(Array.append [|ch|] c), binenv
          | Phrase(c1), Phrase(c2) -> Phrase(Array.append c1 c2), binenv
          | _ as f, (_ as s) -> 
              raise (Illegal_operation((c_type_str f) ^ "+" ^ (c_type_str s)))
        )
        | Sub -> 
            let lit = (
              match lit1, lit2 with
                  Int(x), Int(y) -> Int(x-y)
                | Pitch(letr,oct), Int(x) ->
                    let rcomp = letr - x in
                    if rcomp < 0 then 
                      let posval = rcomp * -1 in
                      let mdls = posval mod 12 in
                      let submod = 12 - mdls in
                      let oreduce = posval/12 in
                      let oret = oct - (oreduce+1) in
                      if oret < 0
                      then Pitch(0,0)
                      else Pitch(submod,oret)
                    else Pitch(rcomp, oct)
                | Dur(num,den) as d, Int(x) -> 
                    dur_sub_abs (d, Dur(x,1))
                | Pitch(l1,o1) as p1, (Pitch(l2,o2) as p2) ->
                    let i1 = pitch_to_int p1 in
                    let i2 = pitch_to_int p2 in
                    Int(i1 - i2)
                | Chord(p_array,d), (Pitch(l,o) as p) ->
                    let pitch_list = Array.fold_left
                      (fun plist next ->
                        if next = p then plist else next :: plist
                      ) [] p_array
                    in let pitches = Array.of_list pitch_list
                    in Chord(pitches,d)
                | Dur(_,_) as d1, (Dur(_,_) as d2) ->
                    dur_sub_abs (d1,d2)
                | Note(p,d1), (Dur(_,_) as d2) -> 
                    Note(p, (dur_sub_abs (d1,d2)))
                | Chord(p,d1), (Dur(_,_) as d2) ->
                    Chord(p, (dur_sub_abs (d1,d2)))
                | _ as f, (_ as s) ->
                    raise (Illegal_operation((c_type_str f) ^ "-" ^ (c_type_str s)))
            ) in lit, binenv
        | Mul -> 
            let lit = (
              match lit1, lit2 with
                  Int(x), Int(y) -> Int(x*y)
                | Dur(n,d), Int(x) 
                | Int(x), Dur(n,d) -> 
                    let rn = n*x in
                    let g = gcd rn d in
                    Dur(rn/g,d/g)
                | Dur(n1,d1), Dur(n2,d2) -> 
                    let n = n1 * n2 in
                    let d = d1 * d2 in
                    let g = gcd n d in
                    Dur(n/g,d/g)
                | _ as f, (_ as s) ->
                    raise (Illegal_operation((c_type_str f) ^ "*" ^ (c_type_str s)))
            ) in lit, binenv
        | Div -> (
            match (lit1, lit2) with
                Int(x), Int(y)            -> Dur(x, y), binenv                                   (* int / int *)
              | Dur(n,d), Int(i)          -> dur_divide (Dur(n,d), Int(i)), binenv               (* dur / int *)
              | Note(p, d), Int(i)        -> Note( p, dur_divide (d, Int(i)) ), binenv           (* note / int *)
              | Chord(p, d), Int(i)       -> Chord( p, dur_divide (d, Int(i)) ), binenv          (* chord / int *)
              | Int(x), Dur(n,d)          -> dur_divide (Int(x), Dur(n,d)), binenv               (* int / dur *)
              | Dur(n1,d1), Dur(n2,d2)    -> dur_divide (Dur(n1,d1), Dur(n2,d2)), binenv         (* dur / dur *)
              | Note(p, dur), Dur(n,d)    -> Note( p, dur_divide (dur, Dur(n,d)) ), binenv       (* note / dur *)
              | Chord(p, chdur), Dur(n,d) -> Chord( p, dur_divide (chdur, Dur(n,d)) ), binenv    (* chord / dur *)
              | Dur(n,d), Note(p, dur)    -> Note(p, dur_divide (Dur(n,d), dur)), binenv         (* dur / note *)
              | Dur(n,d), Chord(p, chdur) -> Chord( p, dur_divide (chdur, Dur(n,d)) ), binenv    (* dur / chord *)
              | _ as f, (_ as s)
                -> raise (Illegal_operation((c_type_str f) ^ "/" ^ (c_type_str s)))
          )
        | Layer -> (
          match (lit1, lit2) with 
            (Dur(n,d1) as d), (Pitch(l,o) as p)
          | (Pitch(l,o) as p), (Dur(n,d1) as d) -> Note(p, d), binenv
          (* TODO: array of pitches and array of durations (* Done? *)*)
          | Array(SPitch, p), Array(SDur, d)
          | Array(SDur, d), Array(SPitch, p) -> (
            if (Array.length p) != (Array.length d) then
              raise (Illegal_operation ("Error: pitch[] and dur[] must be same length"))
            else
              let ph = Array.mapi (fun i e -> Chord([|e|], d.(i))) p in
              Phrase(ph), binenv
          )
          | (Pitch(l,o) as pitch), Note(p,d)
          | Note(p,d), (Pitch(l,o) as pitch) -> 
              Chord([| p; pitch |], d), binenv
          | (Pitch(l,o) as pitch), Chord(p,d)
          | Chord(p,d), (Pitch(l,o) as pitch) -> 
              Chord(Array.append p [| pitch |], d), binenv
          | Note(p1,d1), (Chord(p_array,d2) as c) -> 
              Score([| Phrase([|Chord([|p1|],d1)|]); Phrase([|c|]) |]), binenv
          | (Chord(p_array,d2) as c), Note(p1,d1) -> 
              Score([| Phrase([|c|]); Phrase([|Chord([|p1|],d1)|]) |]), binenv
          | (Chord(p1,d1) as c1), (Chord(p2,d2) as c2) ->
              Score([| Phrase([|c1|]); Phrase([|c2|]) |]), binenv
          | (Chord(p,d) as c), (Phrase(s) as ph) -> 
              Score([| Phrase([|c|]); ph |]), binenv
          | (Phrase(s) as ph), (Chord(p,d) as c) ->
              Score([| ph; Phrase([|c|]) |]), binenv
          | (Phrase(c1) as p1), (Phrase(c2) as p2) -> 
              Score([| p1; p2 |]), binenv
          | (Phrase(c) as p), Score(ph)
          | Score(ph), (Phrase(c) as p) -> Score(Array.append ph [|p|]), binenv
          | _ as f, (_ as s) -> 
              raise (Illegal_operation((c_type_str f) ^ "^" ^ (c_type_str s)))
      )
        | Arrcat -> (
          match (lit1, lit2) with
            Array(t1, d1), Array(t2, d2) -> 
              if t1 = t2 then Array(t1, Array.append d1 d2), binenv
              else raise (Illegal_operation("Error: cannot concatenate arrays of two types"))
          | Array(t, d), x ->
              if c_to_s_type x = t then Array(t, Array.append d [|x|]), binenv
              else raise (Illegal_operation("Error: type mismatch"))
          | x, Array(t,d) ->
              if c_to_s_type x = t then Array(t, Array.append [|x|] d), binenv
              else raise (Illegal_operation("Error: type mismatch"))
          | x, y ->
              if c_to_s_type x = c_to_s_type y then 
                Array(c_to_s_type x, [| x; y |]), binenv
              else raise(Illegal_operation("Error: type mismatch"))
          )
        | Rep -> 
            let lit = (
              match lit1, lit2 with
                  _ as nonint, Int(x)
                | Int(x), (_ as nonint) -> 
                    let arr_type = c_to_s_type nonint
                    in let arr_data = Array.make x nonint
                    in Array(arr_type, arr_data)
                | _ as f, (_ as s) -> 
                    raise (Illegal_operation(
                      "** must be used with at least 1 int, here used with "
                      ^ (c_type_str f) ^ " and " ^ (c_type_str s)
                    ))
            ) in lit, binenv
        | Eq -> ( match lit1, lit2 with
              Int(x), Int(y) ->
                let r = if x = y then 1 else 0 in Int(r),binenv
            | Dur(n1,d1), Dur(n2,d2) ->
                let r = if n1 = n2 && d1 = d2 then 1 else 0
                in Int(r),binenv
            | Pitch(l1,o1), Pitch(l2,o2) ->
                let r = if l1 = l2
                  && o1 = o2
                then 1 else 0
                in Int(r),binenv
            | Note(p1,d1), Note(p2,d2) ->
                let r = if p1 = p2 && d1 = d2 then 1 else 0
                in Int(r),binenv
            | _ as f, (_ as s) ->
                raise (Illegal_operation((c_type_str f) ^ "==" ^ (c_type_str s)))
          )
        | Neq -> ( match lit1, lit2 with
              Int(x), Int(y) ->
                let r = if x = y then 0 else 1 in Int(r),binenv
            | Dur(n1,d1), Dur(n2,d2) ->
                let r = if n1 = n2 && d1 = d2 then 0 else 1
                in Int(r),binenv
            | Pitch(l1,o1), Pitch(l2,o2) ->
                let r = if l1 = l2
                  && o1 = o2
                then 0 else 1
                in Int(r),binenv
            | Note(p1,d1), Note(p2,d2) ->
                let r = if p1 = p2 && d1 = d2 then 0 else 1
                in Int(r),binenv
            | _ as f, (_ as s) ->
                raise (Illegal_operation((c_type_str f) ^ "!=" ^ (c_type_str s)))
          )
        | Gt -> ( match lit1, lit2 with
              Int(x), Int(y) ->
                let r = if x > y then 1 else 0 in Int(r),binenv
            | Dur(n1,d1) as dur1, (Dur(n2,d2) as dur2) ->
                let ctype = dur_sub (dur1,dur2) in
                ( match ctype with
                    Dur(sn,sd) -> let r = if sn > 0 then 1 else 0 in
                      Int(r),binenv
                  | _ -> raise (Illegal_operation("Problem with dur_sub"))
                )
            | Pitch(p1,o1) as pitch1, (Pitch(p2,o2) as pitch2) ->
                let i1 = pitch_to_int pitch1 in
                let i2 = pitch_to_int pitch2 in
                let r = if i1 > i2 then 1 else 0 in
                Int(r),binenv
            | _ as f, (_ as s) ->
                raise (Illegal_operation((c_type_str f) ^ ">" ^ (c_type_str s)))                
          )
        | Gte -> ( match lit1, lit2 with
              Int(x), Int(y) ->
                let r = if x >= y then 1 else 0 in Int(r),binenv
            | Dur(n1,d1) as dur1, (Dur(n2,d2) as dur2) ->
                let ctype = dur_sub (dur1,dur2) in
                ( match ctype with
                    Dur(sn,sd) -> let r = if sn >= 0 then 1 else 0 in
                      Int(r),binenv
                  | _ -> raise (Illegal_operation("Problem with dur_sub"))
                )
            | Pitch(p1,o1) as pitch1, (Pitch(p2,o2) as pitch2) ->
                let i1 = pitch_to_int pitch1 in
                let i2 = pitch_to_int pitch2 in
                let r = if i1 >= i2 then 1 else 0 in
                Int(r),binenv
            | _ as f, (_ as s) ->
                raise (Illegal_operation((c_type_str f) ^ ">=" ^ (c_type_str s)))                
          )
        | Lt -> ( match lit1, lit2 with
              Int(x), Int(y) ->
                let r = if x < y then 1 else 0 in Int(r),binenv
            | Dur(n1,d1) as dur1, (Dur(n2,d2) as dur2) ->
                let ctype = dur_sub (dur1,dur2) in
                ( match ctype with
                    Dur(sn,sd) -> let r = if sn < 0 then 1 else 0 in
                      Int(r),binenv
                  | _ -> raise (Illegal_operation("Problem with dur_sub"))
                )
            | Pitch(p1,o1) as pitch1, (Pitch(p2,o2) as pitch2) ->
                let i1 = pitch_to_int pitch1 in
                let i2 = pitch_to_int pitch2 in
                let r = if i1 < i2 then 1 else 0 in
                Int(r),binenv
            | _ as f, (_ as s) ->
                raise (Illegal_operation((c_type_str f) ^ "<" ^ (c_type_str s)))                
          )
        | Lte -> ( match lit1, lit2 with
              Int(x), Int(y) ->
                let r = if x <= y then 1 else 0 in Int(r),binenv
            | Dur(n1,d1) as dur1, (Dur(n2,d2) as dur2) ->
                let ctype = dur_sub (dur1,dur2) in
                ( match ctype with
                    Dur(sn,sd) -> let r = if sn <= 0 then 1 else 0 in
                      Int(r),binenv
                  | _ -> raise (Illegal_operation("Problem with dur_sub"))
                )
            | Pitch(p1,o1) as pitch1, (Pitch(p2,o2) as pitch2) ->
                let i1 = pitch_to_int pitch1 in
                let i2 = pitch_to_int pitch2 in
                let r = if i1 <= i2 then 1 else 0 in
                Int(r),binenv
            | _ as f, (_ as s) ->
                raise (Illegal_operation((c_type_str f) ^ "<=" ^ (c_type_str s)))                
          )
    )
    | Noexpr -> Int(1), expenv
    
    (* Processes a single statement; returns updated environment *)
    in let rec stmt stenv e = 
      if stenv.return_val <> None then stenv
      else match e with
          Expr e -> let c, renv = expr stenv e in renv
        | Vdecl(t, id) -> (
            try 
              let _ = NameMap.find id stenv.scope.variables
              in raise (Duplicate_name(id ^ " already defined"))
            with Not_found ->
              let typ = type_from_str t in
              let new_entry typ = (
                match typ with
                    SArray -> (Array(arr_type_from_str t, [||]), typ)
                  | _ -> (None, typ)
              )
              in let newval, newtyp = new_entry typ
              in let new_scope = {
                parent = stenv.scope.parent;
                variables = 
                  NameMap.add id (newval, newtyp) stenv.scope.variables
              } in {
                glob_env = stenv.glob_env;
                scope = new_scope;
                return_type = stenv.return_type;
                return_val = stenv.return_val
              }
          )
        | VdeclAsn(s, e) -> (*print_endline ("VdeclAsn");*)
            let newstenv = stmt stenv s in 
            let c, newenv = expr newstenv e in newenv
        | Return(e) -> (*print_endline "Return";*)
            let c, renv = expr stenv e in
            let stype = c_to_s_type c in
            if  stype <> renv.return_type
            then raise (Type_mismatch(
              "Function type " ^ (str_from_type renv.return_type) 
              ^ "does not match function return type " ^ (str_from_type stype)
            ))
            else {
              glob_env = renv.glob_env;
              scope = renv.scope;
              return_type = renv.return_type;
              return_val = c
            }
        | Block(sl) -> (*print_endline "Block";*)
            let block_scope = {
              parent = Some(stenv.scope);
              variables = NameMap.empty
            } in
            let block_env = {
              glob_env = stenv.glob_env;
              scope = block_scope;
              return_type = stenv.return_type;
              return_val = stenv.return_val
            } in
            let post_block_env = List.fold_left stmt block_env sl in
            let post_block_scope = ( match post_block_env.scope.parent with
                Some(parent) -> parent
              | _ -> stenv.scope
            ) in
            {
              glob_env = post_block_env.glob_env;
              scope = post_block_scope;
              return_type = stenv.return_type;
              return_val = stenv.return_val
            }
        | If(e,s_if,s_else) -> 
            let c, ifenv = expr stenv e in (
              match c with
                  Int(x) when x > 0 -> stmt ifenv s_if
                | Int(0) -> stmt ifenv s_else
                | _ -> raise (
                    Type_error("Conditional must be of type int instead of "
                    ^ (c_type_str c))
                  )
            )
        | For(init,cond,iter,st) -> 
            let cinit, initenv = expr stenv init in
            exec_loop initenv cond iter st
    
    and exec_loop env cond iter st =
      let c_cond, condenv = expr env cond in
      match c_cond with
        Int(x) when x > 0 ->
          let loopenv = stmt condenv st in
          let _, iterenv = expr loopenv iter in
          exec_loop iterenv cond iter st
      | Int(0) -> env
      | _ -> 
          raise (Type_error("For loop conditional must be of type int instead of "
            ^ (c_type_str c_cond)))
      
    in stmt env block
)
(* Function execution *)
and exec_fun env fxn args = 
  let params = List.fold_left 
    (
      fun amap (t, id) -> NameMap.add id (None, type_from_str t) amap
    ) NameMap.empty fxn.formals 
  in
  let funscope = {
    parent = None;
    variables = params
  } in
  let empty_fun_env = {
    glob_env = env.glob_env;
    scope = funscope;
    return_type = type_from_str fxn.ftype;
    return_val = None
  } in
  try
    let _, fun_env = List.fold_left2 
      (
        fun (_, fenv) (t, id) c_expr ->
          try
            update_symtab fenv id c_expr (c_to_s_type c_expr)
          with Type_mismatch(s) -> 
            raise (Type_mismatch("Invalid argument to function " ^ fxn.fname ^ ": " ^ id))
      ) (None,empty_fun_env) fxn.formals args
    in
    let post_fun_env = List.fold_left
      (fun nextenv nextstmt ->
        translate nextstmt nextenv
      ) fun_env fxn.body
    in
    post_fun_env (* DO NOT MIX THIS UP WITH THE CALLER'S ENVIRONMENT *)
  with Invalid_argument(s) -> raise (Invalid_argument(fxn.fname ^ ": Wrong number of arguments"))