(*
Functions for semantic analysis
Andrew OReilly, Stephanie Huang
*)

open Ast
open Sast

(* Returns true if pitch is valid *)
let is_valid_pitch (letter, oct) = match letter, oct with
    -1, _ -> true
  | l, o when 0 <= l && l <= 11 && 0 <= o && o <= 8 -> true
  | _, _ -> false

let raise_pitch n p = match n, p with
    Int(x), Pitch(l,o) -> (
      let o = o + (l + x) / 12 in
      let l = (l + x) mod 12 in
      Pitch(l,o)
    )
  | _, _ -> raise (Invalid_type("raise_pitch: invalid type"))

(* Convert a pitch literal to an integer value *)
let pitch_to_int p = 
  match p with Pitch(l, o) -> 9 + o * 12 + l
    | _ -> raise (Invalid_type("pitch_to_int: invalid type"))
   

(* Matches a string with correspondin s_type *)
let type_from_str t = match t with
    "int" -> Sast.SInt
  | "pitch" -> Sast.SPitch
  | "dur" -> Sast.SDur
  | "note" -> Sast.SNote
  | "chord" -> Sast.SChord
  | "phrase" -> Sast.SPhrase
  | "score" -> Sast.SScore
  | "int[]" | "pitch[]" | "dur[]" | "note[]" | "chord[]" | "phrase[]" 
    | "score[]" -> Sast.SArray
  | _ -> raise (Invalid_type ("\""^t^"\""^" is an invalid type"))

(* get components of Array type *)
let get_arr_type a = match a with Array(typ, dat) -> typ
  | _ -> raise (Invalid_type ("Get get array type from non-array"))
  
let get_arr_data a = match a with Array(typ, dat) -> dat
  | _ -> raise (Invalid_type ("Get get array data from non-array"))

(* Check that element and array type match up *)
let is_valid_elem elem arr_type = 
  if ((c_to_s_type elem) = arr_type) then true else false

(* Gets type of array from string *)
let arr_type_from_str t = 
  let typ = 
    if String.contains t '[' then 
      if String.contains t ']' then String.sub t 0 (String.length t - 2)
      else raise (Invalid_type ("\""^t^"\""^" is an invalid array type"))
    else raise (Invalid_type ("\""^t^"\""^" is an invalid array type"))
  in type_from_str typ 

let str_from_type = function
    Sast.SInt -> "int"
  | Sast.SPitch -> "pitch"
  | Sast.SDur -> "dur" 
  | Sast.SNote -> "note" 
  | Sast.SChord -> "chord" 
  | Sast.SPhrase -> "phrase" 
  | Sast.SScore -> "score"
  | Sast.SArray -> "array"
  | None -> "none"
  
(* For reducing fractions *)
let rec gcd a_arg b_arg =
  let a = if a_arg < 0 then a_arg * -1 else a_arg in
  let b = if b_arg < 0 then b_arg * -1 else b_arg in
  if a = 0 then b
  else if b = 0 then a
  else if a = b then
    a
  else if a > b then
    gcd (a - b) b
  else
    gcd a (b - a)

(* Handles cases of dur / int, int / dur, dur / dur *)
let dur_divide (a1, a2) = 
  let n, d = match a1, a2 with
      Dur(dn,dd), Int(i)         -> dn, dd*i
    | Int(i), Dur(dn,dd)         -> dd*i, dn
    | Dur(dn1,dd1), Dur(dn2,dd2) -> dn1*dd2, dd1*dn2
    | _,_ -> 0,0
  in 
  let g = gcd n d in
  Dur(n / g, d / g)

(* Duration subtraction - returns negative values *)
let dur_sub = function
    Dur(n1,d1), Dur(n2,d2) ->
      let n1sub = d2 * n1 in
      let d1sub = d2 * d1 in
      let n2sub = d1 * n2 in
      let n = n1sub - n2sub in
      let g = gcd n d1sub in
      Dur(n / g, d1sub / g)
  | _,_ -> raise (Invalid_type("Error: invalid type. Dur expected."))

(* Same as dur_sub, but returns absolute value *)
let dur_sub_abs = function
    Dur(_,_) as d1, (Dur(_,_) as d2) ->
      ( match dur_sub (d1,d2) with
          Dur(n,d) -> Dur(Pervasives.abs n, d)
        | _ -> raise (Invalid_type("This should never happen"))
      )
  | _,_ -> raise (Invalid_type("Error: invalid type. Dur expected."))
      
(* Add durations *)
let dur_add d1 d2 = match d1, d2 with
    Dur(n1,d1), Dur(n2,d2) -> (
      let n = n1 * d2 + n2 * d1 in
      let d = d1 * d2 in
      let g = gcd n d in
      Dur(n / g, d / g)
    )
  | _, _ -> raise (Invalid_type("Error: invalid type. Dur expected."))
